#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <io.h>
#include <winsock2.h>

typedef int socklen_t;
typedef int ssize_t;
#define MSG_TRUNC 0
#define fcntl(s, k, v) (0)
#define SIOCSIFFLAGS CTL_CODE(0x8102, 0x804, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define VNETBIND  CTL_CODE(0x8102, 0x811, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define VNETRECV  CTL_CODE(0x8102, 0x80B, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IFF_UP        0x001
#define IFF_BROADCAST 0x002
#define IFF_PROMISC   0x008
#define IFF_MULTICAST 0x010
#define IFF_RUNNING   0x020
#define IFF_ALLMULTI  0x200

static const char* mystrerror(int errorcode) {
	static char buf[4096];

	if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, 0, errorcode, 0, buf, sizeof(buf), NULL) == 0) {
		return "Unknown error";
	}
	return buf;
}

int initSockets(void) {
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;

	wVersionRequested = MAKEWORD( 2, 0 );

	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		fprintf(stderr, "Unable to initialize WinSock\n");
		return 21;
	}
	if (LOBYTE( wsaData.wVersion ) != 2 ||
	    HIBYTE( wsaData.wVersion ) != 0 ) {
		fprintf(stderr, "Sorry, this version 2.0 of WinSock is required\n");
		WSACleanup();
		return 22;
	}
	return 0;
}

static int childProc(void) {
	WSAPROTOCOL_INFO pin;
	DWORD rtn;
	SOCKET c2;

	while (1) {
		if (!ReadFile(GetStdHandle(STD_INPUT_HANDLE), &pin, sizeof(pin), &rtn, NULL)) {
			fprintf(stderr, "Cannot read from file: %s\n", mystrerror(GetLastError()));
			return 15;
		}
		printf("Read %u bytes\n", rtn);
		c2 = WSASocket(PF_INET, SOCK_STREAM, IPPROTO_TCP, &pin, 0, 0);
		if (c2 == INVALID_SOCKET) {
			fprintf(stderr, "Cannot create socket duplicate: %s\n", mystrerror(WSAGetLastError()));
			return 15;
		}
		WriteFile((HANDLE)c2, "Hi there\r\n", 10, &rtn, NULL);
		closesocket(c2);
	}
	return 0;
}

static int childClientProc(void) {
	WSAPROTOCOL_INFO pin;
	DWORD rtn;
	SOCKET c2;
	char b[1000];

	if (!ReadFile(GetStdHandle(STD_INPUT_HANDLE), &pin, sizeof(pin), &rtn, NULL)) {
		fprintf(stderr, "Cannot read from file: %s\n", mystrerror(GetLastError()));
		return 15;
	}
	printf("Read %u bytes\n", rtn);
	c2 = WSASocket(PF_INET, SOCK_STREAM, IPPROTO_TCP, &pin, 0, 0);
	if (c2 == INVALID_SOCKET) {
		fprintf(stderr, "Cannot create socket duplicate: %s\n", mystrerror(WSAGetLastError()));
		return 15;
	}
	if (!ReadFile((HANDLE)c2, b, sizeof(b), &rtn, NULL)) {
		fprintf(stderr, "Cannot read from file: %s\n", mystrerror(GetLastError()));
		return 15;
	}
	closesocket(c2);
	printf("Got answer: %u bytes, text: ", rtn);
	fwrite(b, 1, rtn, stdout);
	return 0;
}

static int doClient(SOCKET s, const char* host, HANDLE pip, DWORD pid, HANDLE hp) {
	struct hostent* he;
	struct sockaddr_in sin;
	WSAPROTOCOL_INFO pin;
	DWORD rtn;
	int i;

	he = gethostbyname(host);
	if (!he) {
		fprintf(stderr, "Cannot resolve hostname: %s\n", mystrerror(WSAGetLastError()));
		closesocket(s);
		return 15;
	}
	sin.sin_family = AF_INET;
	memcpy(&sin.sin_addr.s_addr, he->h_addr, 4);
	sin.sin_port = htons(20000);
	connect(s, (struct sockaddr*)&sin, sizeof(sin));
	i = WSADuplicateSocket(s, pid, &pin);
	if (i == SOCKET_ERROR) {
		fprintf(stderr, "Cannot duplicate socket: %s\n", mystrerror(WSAGetLastError()));
		closesocket(s);
		return 15;
	}
	WriteFile(pip, &pin, sizeof(pin), &rtn, NULL);
	closesocket(s);
	CloseHandle(pip);
	/* Wait until child exits... Say 5 seconds*/
	WaitForSingleObject(hp, 5000);
	return 0;
}

int main(int argc, char* argv[]) {
	DWORD pid;
	WSAPROTOCOL_INFO pin;
	SOCKET sock;
	SOCKET chld;
	SOCKET c2;
	DWORD masterL;
	struct sockaddr_in sin;
	int i;
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	HANDLE ho;
	SECURITY_ATTRIBUTES sa;
	int isclnt = 0;

	if (initSockets()) {
		return 21;
	}

	if (argv[1] && !strcmp(argv[1], "chld")) {
		childProc();
		return 99;
	}
	if (argv[1] && !strcmp(argv[1], "cchld")) {
		childClientProc();
		return 99;
	}
	if (argc == 2) {
		isclnt = 1;
	}
	memset(&si, 0, sizeof(si));
	si.cb = sizeof(si);
	si.dwFlags = STARTF_USESTDHANDLES;
	si.hStdInput = 0;
	si.hStdOutput = GetStdHandle(STD_OUTPUT_HANDLE);
	si.hStdError = GetStdHandle(STD_ERROR_HANDLE);

	memset(&sa, 0, sizeof(sa));
	sa.nLength = sizeof(sa);
	sa.bInheritHandle = TRUE;
	CreatePipe(&si.hStdInput, &ho, &sa, 0);
	CreateProcess(argv[0], isclnt ? "cchld cchld cchld cchld" : "chld chld chld chld", NULL, NULL, TRUE, 0, NULL, NULL, &si, &pi);
	CloseHandle(si.hStdInput);
	CloseHandle(pi.hThread);
	if (!isclnt) {
		CloseHandle(pi.hProcess);
	}
	sock = WSASocket(PF_INET, SOCK_STREAM, IPPROTO_TCP, NULL, 0, 0);
	if (sock == INVALID_SOCKET) {
		fprintf(stderr, "Cannot create socket: %s\n", mystrerror(WSAGetLastError()));
		return 15;
	}
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl(INADDR_ANY);
	sin.sin_port = htons(isclnt ? 0 : 20000);
	if (bind(sock, (struct sockaddr*)&sin, sizeof(sin))) {
		fprintf(stderr, "Cannot bind to socket: %s\n", mystrerror(WSAGetLastError()));
		closesocket(sock);
		return 14;
	}
	if (isclnt) {
		return doClient(sock, argv[1], ho, pi.dwProcessId, pi.hProcess);
	}
	listen(sock, 10);
	while (1) {
		chld = WSAAccept(sock, NULL, NULL, NULL, NULL);
		if (chld == INVALID_SOCKET) {
			fprintf(stderr, "Cannot create socket: %s\n", mystrerror(WSAGetLastError()));
			closesocket(sock);
			return 15;
		}
		pid = pi.dwProcessId;
		i = WSADuplicateSocket(chld, pid, &pin);
		if (i == SOCKET_ERROR) {
			fprintf(stderr, "Cannot duplicate socket: %s\n", mystrerror(WSAGetLastError()));
			closesocket(sock);
			return 15;
		}
		WriteFile(ho, &pin, sizeof(pin), &masterL, NULL);
		closesocket(chld);
	}
	CloseHandle(ho);
	closesocket(sock);
	return 0;
}


